@@ -1,6 +1,5 @@
-#!/usr/bin/perl
-
 package Thread::Workers;
+
 use 5.012;
 
 use warnings;
@@ -15,7 +14,7 @@ use Thread::Semaphore;
 use Encode;
 
 our $VERSION;
-$VERSION = '0.02';
+$VERSION = '0.03';
 
 sub new {
     my $class = shift;
@@ -23,7 +22,7 @@ sub new {
     
     $self->{params} = {@_} || {};
     if (!exists($self->{params}{threadcount})) {
-	$self->{params}{threadcount} = 1; # this is the minimum of 1 thread
+	$self->{params}{threadcount} = 2; # this is the minimum of 1 thread
     }
     if (!exists($self->{params}{threadinterval})) {
 	$self->{params}{threadinterval} = 5;
@@ -45,6 +44,7 @@ sub _initialize() {
     $self->{_boss_tid} = 0;
     $self->{_boss_lock} = Thread::Semaphore->new(1);
     $self->{_boss_cmd} = 0;
+    $self->{_boss_cmd_lock} = Thread::Semaphore->new(1);
     $self->{_worker_lock} = Thread::Semaphore->new($self->{params}{threadcount});
     $self->{_worker_log_lock} = Thread::Semaphore->new(1);
     $self->{_worker_log} = undef;
@@ -52,15 +52,18 @@ sub _initialize() {
     share($self->{_boss_lock});
     share($self->{_boss_tid});
     share($self->{_boss_cmd});
+    share($self->{_boss_cmd_lock});
     share($self->{_worker_lock});
     share(@{$self->{_worker_log}});
     share($self->{_worker_log_lock});
     
-    for my $thr (1..$self->{params}{threadcount}) {
+    for my $thr (2..$self->{params}{threadcount}) {
 	$self->{_worker_tid}{$thr} = 0;
 	$self->{_worker_cmd}{$thr} = 0;
+	$self->{_worker_cmd_lock}{$thr} = Thread::Semaphore->new(1);
 	share($self->{_worker_cmd}{$thr});
 	share($self->{_worker_tid}{$thr});
+	share($self->{_worker_cmd_lock}{$thr});
     }
 }
 
@@ -76,47 +79,51 @@ sub start_boss {
     }
 }
 
+sub _main_boss_loop {
+    my $self = shift;
+    my $work;
+    $self->{_boss_cmd_lock}->down();
+    $self->{_boss_cmd} = 1;
+    $self->{_boss_cmd_lock}->up();
+    while ($self->{_boss_cmd}) {
+        $work = $self->{boss}{cb}->();
+        # Do work on $item
+        if ($work) { 
+            if (ref($work) eq 'HASH') {
+                my @task_list;
+                foreach my $key (keys %$work) {
+                    my $num = int(decode_utf8($key));
+                    $task_list[$num] = $work->{$key};
+                }
+                for my $tsk (0..$#task_list) {
+                    $self->{_queue}->enqueue($task_list[$tsk]);
+                }
+            } 
+            else {
+                $self->{_queue}->enqueue($work);
+            }
+        };
+        $self->{_worker_log_lock}->down();
+        if ($#{$self->{_work_log}}) {
+            if (exists($self->{boss}{logcb})) {
+                $self->{boss}{logcb}->($self->{_work_log});
+            } 
+            #erase worker results
+            $self->{_work_log} = undef; 
+        }
+        $self->{_worker_log_lock}->up();
+        if ($self->{_boss_cmd}) {
+    	    do {
+                sleep($self->{params}{bossinterval});
+            } until ($self->{_boss_cmd} <= 1);
+        }
+    }
+}
+
 sub _start_boss {
     my $self = shift;
     if (exists($self->{boss}{cb}) and exists($self->{_queue}) and ($self->{_boss_lock}->down_nb())) {
-	$self->{_boss} = threads->create(sub {
-				my $work;
-				my $alive = 1;    #
-                                while ($self->{_boss_cmd}) {
-                            	    $work = $self->{boss}{cb}->();
-                                    # Do work on $item
-                                    if ($work) { 
-                                	if (ref($work) eq 'HASH') {
-                                	    my @task_list;
-                                	    foreach my $key (keys %$work) {
-                                		my $num = int(decode_utf8($key));
-                                		$task_list[$num] = $work->{$key};
-                                	    }
-                                	    for my $tsk (0..$#task_list) {
-                                		$self->{_queue}->enqueue($task_list[$tsk]);
-                                	    }
-                                	} 
-                                	else {
-                                	    $self->{_queue}->enqueue($work);
-                                	}
-                            	    };
-                            	    $self->{_worker_log_lock}->down();
-                            	    if ($#{$self->{_work_log}}) {
-                            	    
-                            		if (exists($self->{boss}{logcb})) {
-                            		    $self->{boss}{logcb}->($self->{_work_log});
-                            		} 
-                            		#erase worker results
-                            		$self->{_work_log} = undef; 
-                            	    }
-                            	    $self->{_worker_log_lock}->up();
-                            	    if ($self->{_boss_cmd}) {
-                            		 do {
-                                	    sleep($self->{params}{bossinterval});
-                            		} until ($self->{_boss_cmd} <= 1);
-                            	    }
-                                };
-        });
+	$self->{_boss} = threads->create(sub { $self->_main_boss_loop } );
         $self->{_boss_tid} = $self->{_boss}->tid;
     }
     else { 
@@ -148,8 +155,11 @@ sub add_worker {
     my $self = shift;
     $self->{params}{threadcount}++;
     $self->{_worker_lock}->up();
+    $self->{_worker_cmd}{$self->{params}{threadcount}} = 0;
+    $self->{_worker_cmd_lock}{$self->{params}{threadcount}} = Thread::Semaphore->new(1);
     share($self->{_worker_tid}{$self->{params}{threadcount}});
     share($self->{_worker_cmd}{$self->{params}{threadcount}});
+    share($self->{_worker_cmd_lock}{$self->{params}{threadcount}});
     $self->_workers_begin(1);
     return 1;
 }
@@ -170,10 +180,12 @@ sub start_workers {
 
 sub stop_workers {
     my $self = shift;
-    for my $thr (1..$self->{params}{threadcount}) {
+    for my $thr (2..$self->{params}{threadcount}) {
+	$self->{_worker_cmd_lock}{$thr}->down();
 	$self->{_worker_cmd}{$thr} = 0;
+	$self->{_worker_cmd_lock}{$thr}->up();
     }
-    for my $thr (1..$self->{params}{threadcount}) {
+    for my $thr (2..$self->{params}{threadcount}) {
 	$self->{_worker}{$thr}->join();
 	$self->{_worker_tid}{$thr} = 0;
 	$self->{_worker_lock}->up();
@@ -193,45 +205,48 @@ sub stop_worker {
     }
 }
 
+sub _worker_main_loop {
+    my $self = shift;
+    my $thr = shift;
+    $self->{_worker_lock}->down();    #
+    $self->{_worker_cmd_lock}{$thr}->down();
+    $self->{_worker_cmd}{$thr} = 1;    #
+    $self->{_worker_cmd_lock}{$thr}->up();
+    WORK: while () {
+        my $work = $self->{_queue}->dequeue_nb();
+        # Do work on $item
+        if ($work) { 
+            my $results = $self->{_worker_cb}->($work);
+            if (defined($results)) {
+                $self->{_worker_log_lock}->down();
+                push @{$self->{_worker_log}}, $work => $results;
+                $self->{_worker_log_lock}->up();
+            }
+        }
+        if ($self->{_worker_cmd}{$thr} > 0) { # we're still in RUN state
+    	    do {
+                sleep($self->{params}{threadinterval});
+            } until ($self->{_worker_cmd}{$thr} <= 1);
+        } 
+        else {			# got a DIE signal
+            last WORK;
+        }
+    }
+}
+
 sub _workers_begin {
     my $self = shift;
     my $single = shift;
-
-    my $start_thread = (keys %{$self->{_worker}});
-    # this bit of ugliness is in case we're just starting a single worker so we get its thr num right.
-
-    if (!defined($start_thread)) { $start_thread = 0 };
+    my $start_offset = 2;
     if (defined($single)) {
-	$start_thread++;
+	$start_offset = $self->{params}{threadcount};  # if we're adding a single
     }
     #setup signal watchers, check, sleep for interval - time i took to run
-    for my $thr (1+$start_thread..$self->{params}{threadcount}) {
-	$self->{_worker}{$thr} = threads->create(sub {
-				my $work; 
-				$self->{_worker_lock}->down();    #
-				$self->{_worker_cmd}{$thr} = 1;    #
-                                while ($self->{_worker_cmd}{$thr}) {
-                            	    $work = $self->{_queue}->dequeue_nb();
-                                    # Do work on $item
-                                    if ($work) { 
-                                	my $results = $self->{_worker_cb}->($work);
-                                	if (defined($results)) {
-                                	    $self->{_worker_log_lock}->down();
-                                	    push @{$self->{_worker_log}}, $work => $results;
-                                	    $self->{_worker_log_lock}->up();
-                                	}
-                            	    }
-                                    if ($self->{_worker_cmd}{$thr}) { # we're still in RUN state
-                                	do {
-                                	    sleep($self->{params}{threadinterval});
-                                	} until ($self->{_worker_cmd}{$thr} <= 1);
-
-                            	    } 
-                            	    else {			# got a DIE signal
-                            		last;
-                            	    }
-                                }
-        });
+    for my $thr ($start_offset..$self->{params}{threadcount}) {
+	$self->{_worker}{$thr} = threads->create( sub {
+				    my $tid = threads->tid;
+				    $self->_worker_main_loop($tid);
+	});
         $self->{_worker_tid}{$thr} = $self->{_worker}{$thr}->tid;
     };
 }
@@ -253,9 +268,11 @@ sub stop_boss {
 
 sub sleep_workers {
     my $self = shift;
-    for my $thr (1..$self->{params}{threadcount}+1) {
-	if ($self->{_worker_cmd} > 0) {
+    for my $thr (2..$self->{params}{threadcount}) {
+	if ($self->{_worker_cmd}{$thr} > 0) {
+	    $self->{_worker_cmd_lock}{$thr}->down();
 	    $self->{_worker_cmd}{$thr} = 2;
+	    $self->{_worker_cmd_lock}{$thr}->up();
 	}
     }
     return 1;
@@ -263,9 +280,11 @@ sub sleep_workers {
 
 sub wake_workers {
     my $self = shift;
-    for my $thr (1..$self->{params}{threadcount}) {
+    for my $thr (2..$self->{params}{threadcount}) {
 	if ($self->{_worker_cmd} > 1) {
+	    $self->{_worker_cmd_lock}{$thr}->down();
 	    $self->{_worker_cmd}{$thr} = 1;
+	    $self->{_worker_cmd_lock}{$thr}->up();
 	}
     }
     return 1;
@@ -278,11 +297,11 @@ sub stop_finish_work {
 	sleep(2);
     }
     #send a command to stop all the workers so they pick it up and we don't have to wait through cycles
-    for my $thr (1..$self->{params}{threadcount}) {
+    for my $thr (2..$self->{params}{threadcount}) {
 	$self->{_worker_cmd}{$thr} = 0;
     }
     #join the threads and give back the semaphore tokens
-    for my $thr (1..$self->{params}{threadcount}) {
+    for my $thr (2..$self->{params}{threadcount}) {
 	$self->{_worker}{$thr}->join();
 	$self->{_worker_tid}{$thr} = 0;
 	$self->{_worker_lock}->up();
@@ -327,7 +346,10 @@ works as advertised now.
   $pool->set_worker_work_cb(\&function_does_work);
   $pool->start_boss();
   $pool->start_workers();
-
+  $pool->add_worker();
+  $pool->sleep_workers();
+  $pool->wake_workers();
+  
   #internal control loops
   # we have orders to increase the load! add 500 workers
   for (1..500) { 
